import greenfoot.*;  // (World, Actor, GreenfootImage, Greenfoot and MouseInfo)

/**
 * This more advanced player includes demonstrations of:
 * - Custom Colliders
 * - Move advanced movement
 * - Changing the spritesheet (equipping and removing pickaxe)
 * - Multiple animation states
 * 
 * @author Jordan Cohen 
 * @since 0.8.0
 * @version 0.8.1
 */
public class ComplexPlayer extends AnimatedCharacter
{
    private MouseInfo m;
    private String key;

    private int walkSpeed;
    private int walkAnimSpeed;

    private int runSpeed;
    private int runAnimSpeed;

    private boolean attacking;

    private GreenfootImage spriteSheet;
    private GreenfootImage pickaxe;
    
    private Animation playerWalk, playerSwing;

    private int moveX, moveY;
    private int xOffset, yOffset; // direction for attacking

    public ComplexPlayer () {

        // =-== Set variables for SPEED ==-=

        // Set two different speeds so that I can switch between them later. 
        walkSpeed = 60; // pixels to move per SECOND (not per act)
        walkAnimSpeed = 20; // number of animations frames per SECOND 
        runSpeed = 160; // RUN movement speed
        runAnimSpeed = 32; // RUN animation speed

        // set the speed to start
        changeSpeed(walkSpeed, walkAnimSpeed);

        //  =-== SETUP ANIMATIONS ==-=

        // Cache the pickaxe image, since I intend to add and remove it,
        // and I dont' want to have to load the file from disk each time:
        pickaxe = new GreenfootImage("pickaxe.png");
        
        // First, create my sprite sheets
        setLayer(0, new GreenfootImage("dark.png"));
        setLayer(1, new GreenfootImage("plate_chest_male.png"));
        setLayer(2, new GreenfootImage("plate_pants_male.png"));
        setLayer(3, new GreenfootImage("metal_gloves_male.png"));
        setLayer(4, pickaxe); // if you are going to add/remove objects more than once, load into an object to avoid extra disk reads

        // use Animation methods to build two animations:
        // a walking animation for our primary animation / movement
        animations.put ("move", Animation.createWalkingAnimation(getSpriteSheet()));
        // a swing animation for swinging a weapon or tool
        animations.put ("swing", Animation.createAnimation(getSpriteSheet(), 13, 4, 6, 64, 64));

        // Trim the Animations, so that their hitbox isn't larger than them.
        // This is a manual process for now. You just specify how much to 
        // trim from each of the 4 sides (see API)
        //Animation.trim(animations.get("move"), 12, 12, 12, 2);
        //Animation.trim(animations.get("swing"), 12, 12, 12, 2);

        // Always set a primary animation. This is the default animation. After a terminal
        // animation runs, such as swinging a pickaxe, the animator will default to this.
        primaryAnimation = animations.get("move");
        // Start the player off facing downward
        direction = Direction.DOWN;

        // create a custom collider (width, height, xOffset, yOffset)
        setCollider (28, 40, 0, 6);
        enableCollision();
        
        

        // For the starting image, grab the 0th frame from the current
        // facing dirction
        setImage(primaryAnimation.getOneImage(direction, 0));
    }

    /**
     * Act - do whatever the CostumedPlayer wants to do. This method is called whenever
     * the 'Act' or 'Run' button gets pressed in the environment.
     */
    public void act() 
    {
        // Keyboard iput for pressed keys and Mouse Info:
        // Note: In a real project, consider putting getKey() and getMouseInfo() 
        // in your World, because there can only be ONE getKey() and getMouseInfo() 
        // call per act for the whole simulation, and it's not usually ideal 
        // to put it in your Player class. This is intended as a demonstration.
        m = Greenfoot.getMouseInfo();
        key = Greenfoot.getKey();

        if (key != null){ 
            if (key.equals("space")){
                // A terminal animation will run and then return to the original animation
                // when it's finished running. 
                runTerminalAnimation ("swing", false, false, direction, 1);
                attacking = true;               
            }
            else if (key.equals("s")){
                setLayer(4, null);
                refresh(primaryAnimation);
            } else if (key.equals ("a")){
                setLayer(4, pickaxe);
                refresh(primaryAnimation);
            }
        }

        if (attacking){
            attack();
        }

        // Mouse control
        if (Greenfoot.mouseClicked(null)){
            if (m != null){
                if (m.getButton()==1){
                    moveTowards(m.getX(), m.getY());
                }
            }
        }

        if (Greenfoot.isKeyDown("shift")){
            changeSpeed(runSpeed, runAnimSpeed);
        } else {
            changeSpeed(walkSpeed, walkAnimSpeed);
        }

        // Each act, movement and idle state are reset
        moveX = 0; 
        moveY = 0;

        
        // For each key...
        if (Greenfoot.isKeyDown("right")){
            moveX = 1; // set direction
        }
        if (Greenfoot.isKeyDown("left")){
            moveX = -1;
        }
        if (moveX == 0){ // prevent diagonal movement - only check for Y if nothing is moving on X
            if (Greenfoot.isKeyDown("up")){
                moveY = -1;
            }
            if (Greenfoot.isKeyDown("down")){
                moveY = 1;
            }
        }

        if (moveX != 0 || moveY != 0){
            // set sides/directions for attacking. -1, 0 or 1 for each axis to represent
            // which direction the player is facing, to be used in collision detection
            xOffset = moveX;
            yOffset = moveY;
            moveInDirection (moveX, moveY);
        } else if (key == null) {
            if (!isMoving())
                stopMoving();
        }

        Actor a = checkCollision ();
        if (a != null){
            //stepBack(a); // step back from the specified object
            stepBack();
            stopMoving();
        }

        // Call superclass' act() to perform animations and movement
        super.act(); // <<== DO NOT REMOVE: MUST BE INCLUDED 
        //      IN ALL ANIMATED CHARACTERS
    }    

    private void attack (){
        // Look for an enemy 1 pixel away in the direction that I'm facing
        Enemy e = (Enemy)getOneObjectAtOffset(xOffset * (getImage().getWidth()/2 + 1), yOffset * (getImage().getWidth()/2 + 1), Enemy.class);
        if (e != null){
            e.killMe();
        }
    }

    // OPTIONAL include this method if you want to be notified when an animation is complete. 
    protected void terminalAnimationComplete(){
        attacking = false;
        //currentAnimation = primaryAnimation;
    }
}
